import pandas as pd
import re
import os
from pathlib import Path
module_path = Path(__file__).parent.parent.absolute()


AMBIGRAM_BALANCED_PATH = os.path.join(module_path, "data/ambigram_nc5_len7.csv")
AMBIGRAM_ACCURATE_PATH = os.path.join(module_path, "data/ambigram_nc1_len10.csv")

extract_mapping = lambda df: df[~df.REDUNDANT][["MASKED", "NGRAM"]].values

ambigrams_balanced = extract_mapping(pd.read_csv(AMBIGRAM_BALANCED_PATH))
ambigrams_accurate = extract_mapping(pd.read_csv(AMBIGRAM_ACCURATE_PATH))


diphthong_mapping = [
    ("tsch", "ჭ"),
    ("tch", "ჭ"),
    ("ch", "ჩ"),
    ("sh", "შ"),
    ("dz", "ძ"),
    ("kh", "ხ"),
    ("th", "თ"),
    ("zh", "ჟ"),
    ("jh", "ჟ"),
    ("ph", "ფ"),
    ("gh", "ღ"),
    ("ts", "ც"),
]

common_errors = {
    "t" : ["ტ", "თ"],
    "p" : ["ფ", "პ"],
    "c" : ["ც", "წ"],
    "k" : ["ქ", "კ"],
    "j" : ["ჯ", "ჟ"],
    # "g" : ["გ", "ღ"],
}

letter_mapping = [    
    ('a', 'ა'),
    ('A', 'ა'),
    ('b', 'ბ'),
    ('B', 'ბ'),
    ('c', 'ც'),
    ('C', 'ჩ'),
    ('d', 'დ'),
    ('D', 'დ'),
    ('e', 'ე'),
    ('E', 'ე'),
    ('f', 'ფ'),
    ('F', 'ფ'),
    ('g', 'გ'),
    ('G', 'გ'),
    ('h', 'ჰ'),
    ('H', 'ჰ'),
    ('i', 'ი'),
    ('I', 'ი'),
    ('j', 'ჯ'),
    ('J', 'ჯ'),
    ('k', 'კ'),
    ('K', 'კ'),
    ('l', 'ლ'),
    ('L', 'ლ'),
    ('m', 'მ'),
    ('M', 'მ'),
    ('n', 'ნ'),
    ('N', 'ნ'),
    ('o', 'ო'),
    ('O', 'ო'),
    ('p', 'პ'),
    ('P', 'პ'),
    ('q', 'ქ'),
    ('Q', 'ქ'),
    ('r', 'რ'),
    ('R', 'ღ'),
    ('s', 'ს'),
    ('S', 'შ'),
    ('t', 'ტ'),
    ('T', 'თ'),
    ('u', 'უ'),
    ('U', 'უ'),
    ('v', 'ვ'),
    ('V', 'ვ'),
    ('w', 'წ'),
    ('W', 'ჭ'),
    ('x', 'ხ'),
    ('X', 'ხ'),
    ('y', 'ყ'),
    ('Y', 'ყ'),
    ('z', 'ზ'),
    ('Z', 'ძ')
]


def _decapitalise(text):
    sentences = re.split(r'([.!?]\s*)', text)
    
    def lowercase_first_word(sentence):
        words = sentence.split(" ", 1)
        words[0] = words[0].lower()
        return " ".join(words)
    
    processed_sentences = [lowercase_first_word(sentence) for sentence in sentences[::2]]
    result = ''.join([
        processed_sentences[i] + (sentences[i * 2 + 1] if i * 2 + 1 < len(sentences) else '')
        for i in range(len(processed_sentences))
    ])

    return result


def _transliterate_diphthongs(sentence):
    for dfrom, dto in diphthong_mapping:
        sentence = sentence.replace(dfrom, dto)

    return sentence


def _transliterate_letters(sentence, even_ambiguous=True):
    for mfrom, mto in letter_mapping:
        if not even_ambiguous and mfrom in common_errors:
            continue

        sentence = sentence.replace(mfrom, mto)

    return sentence


def _chevron_wrap(sentence):
    sentence = re.sub(r'\b(?:\w+-\w+|\w+)\b', lambda match: '<' + match.group() + '>', sentence)

    return sentence

def _chevron_unwrap(sentence):
    sentence = re.sub(r"<(\w+)>", r"\1", sentence)

    return sentence

def _transliterate_ngrams(sentence, ambigrams):
    for nfrom, nto in ambigrams:
        if nfrom in sentence:
            sentence = sentence.replace(nfrom, nto)

    return sentence


def georgianise_fast(sentence):
    """
    Fastest implementation of proper Georgianisation.
    """

    sentence = _transliterate_diphthongs(sentence)
    sentence = _transliterate_letters(sentence)

    return sentence


def georgianise_balanced(sentence):
    """
    Smarter than fastest, faster than Smartest. Uses 10K tokens.
    """
    
    sentence = _decapitalise(sentence)
    sentence = _chevron_wrap(sentence)
    sentence = _transliterate_diphthongs(sentence)
    sentence = _transliterate_letters(sentence, even_ambiguous=False)
    sentence = _transliterate_ngrams(sentence, ambigrams=ambigrams_balanced)
    sentence = _chevron_unwrap(sentence)

    return sentence    


def georgianise_accurate(sentence):
    """
    Most accurate implementation of Georgianisation without applying comprehensive million row wordlist in brute force. Uses 50K tokens.
    """

    sentence = _decapitalise(sentence)
    sentence = _chevron_wrap(sentence)
    sentence = _transliterate_diphthongs(sentence)
    sentence = _transliterate_letters(sentence, even_ambiguous=False)
    sentence = _transliterate_ngrams(sentence, ambigrams=ambigrams_accurate)
    sentence = _chevron_unwrap(sentence)

    return sentence    


def georgianise_custom(sentence, ambigrams_custom):
    """
    Custom Georgianisation for performance evaluation.
    """
    
    sentence = _chevron_wrap(sentence)
    sentence = _transliterate_diphthongs(sentence)
    sentence = _transliterate_letters(sentence, even_ambiguous=False)
    sentence = _transliterate_ngrams(sentence, ambigrams=ambigrams_custom)
    sentence = _chevron_unwrap(sentence)

    return sentence       
    
    

def georgianise(sentence, mode="accurate"):
    """
    Redeem others' qwerty sins.
    """

    if mode == "accurate":
        return georgianise_accurate(sentence)
    
    elif mode == "balanced":
        return georgianise_balanced(sentence)
    
    return georgianise_fast(sentence)

    

def latinise(sentence):
    """
    Blasphemous utility.
    """
    
    for latin, georgian_letters in common_errors.items():
        for letter in georgian_letters:
            sentence = sentence.replace(letter, latin)

    for dto, dfrom in diphthong_mapping:
        sentence = sentence.replace(dfrom, dto)
    
    for mto, mfrom in letter_mapping:
        sentence = sentence.replace(mfrom, mto)
    
    return sentence





import pandas as pd
import os
import re
from pathlib import Path

module_path = Path(__file__).parent.parent.absolute()

cdf = pd.read_csv(os.path.join(module_path, "data/georgian_contractions.csv"))
cmap = cdf.set_index('CONTRACTION')[['EXPANSION']].to_dict('dict')['EXPANSION']

def expand(word):
    # Just a wrapper around contractions map
    if word in cmap:
        return cmap[word]

    return word


def expand_text(text):
    # Look up contractions sequentially and replace them with expansions
    while True:
        match = re.search(r'[ა-ჿ]+\.', text)
        if bool(match) == False:
            break

        expansion = expand(text[match.start():match.end()])
        text = text[:match.start()] + expansion + text[match.end():]

    return text




import re 


def nested_tokenize_by_separators(text, first_separator='\n', second_separator=' '):
    text = text.split(first_separator)
    text = [
        [
            token 
            for token in sentence.strip().split(second_separator) 
            if token != ''
        ] 
        for sentence in text
        if sentence != ''
    ]
    return text

def sentence_tokenize(text):
    text = re.sub(r"[^Ⴀ-ჿⴀ-ⴥᲐ-Ჿ\.\?\!]", ' ', text)
    text = re.sub(r"[\?\!]", '.', text)
    text = re.sub(r"\s+", ' ', text)

    # Deals with Mtavruli -> Mkhedruli, Asomtavruli -> Nuskhuri
    # TODO Nuskhuri -> Mkhedruli
    text = text.lower()
    text = nested_tokenize_by_separators(text, '.')

    return text


def paragraph_tokenize(text):
    text = re.sub(r"[^Ⴀ-ჿⴀ-ⴥᲐ-Ჿ\n]", ' ', text)

    # Deals with Mtavruli -> Mkhedruli, Asomtavruli -> Nuskhuri
    # TODO Nuskhuri -> Mkhedruli
    text = text.lower()
    text = nested_tokenize_by_separators(text, '\n')
    
    return text
    


def word_tokenize(text):
    text = re.sub(r"[^Ⴀ-ჿⴀ-ⴥᲐ-Ჿ]", ' ', text)
    text = re.sub(r"\s+", ' ', text)

    # Deals with Mtavruli -> Mkhedruli, Asomtavruli -> Nuskhuri
    # TODO Nuskhuri -> Mkhedruli
    text = text.lower()
    text = [token for token in text.split(' ') if token != '']
    
    return text


def cleanup(text):
    text = re.sub(r"[^Ⴀ-ჿⴀ-ⴥᲐ-Ჿ\.\?\!\,]", ' ', text)
    text = re.sub(r"\s+", ' ', text)

    # Deals with Mtavruli -> Mkhedruli, Asomtavruli -> Nuskhuri
    # TODO Nuskhuri -> Mkhedruli
    text = text.lower()
    return text

# sentence_tokenize(text)
# paragraph_tokenize(text)
# word_tokenize(text)


import hjson
import re
import os

dirname = os.path.dirname(__file__)
datafilepath = os.path.realpath(os.path.join(dirname, "../data/letters.js"))

# Exact copy of the Javascript object from Anbani.JS for easier parallel development down the road
# .js because .json doesn't allow comments that would be parsable with hjson (human json)
data = hjson.loads(
    open(datafilepath).read().replace('export default ', '')
)

script_regex = {
    "mkhedruli": "[ა-ჿ]",
    "mtavruli": "[Ა-Ჿ]",
    "asomtavruli": "[Ⴀ-Ⴥ]",
    "nuskhuri": "[ⴀ-ⴥ]",
    "latin": "[a-zA-Z]",
    "cyrillic": "[А-Яа-я]"
}


# Single entrypoint for both bicameral and unicameral conversion
def convert(text, dir_from, dir_to):
    assert dir_from in [
        'mkhedruli',
        'mtavruli',
        'asomtavruli',
        'nuskhuri',
        'qwerty',
    ]

    if dir_to in [
        "anbanismtavruli",
        "shanidziseuli",
        "tfileliseuli",
        "khutsuri",
    ]:
        return convert_bicameral(text, dir_from, dir_to)
    else:
        return convert_unicameral(text, dir_from, dir_to)


# Source text script automatching, handy for automatic keyboards and etc
def interpret(text, dir_to):
    dir_from = classify_text(text)
    return convert(text, dir_from, dir_to)


# Loose script classifier
# TODO perhaps I should match with Anbani.JS vector matching or smth else entirely
def classify_text(text):
    for key in script_regex.keys():
        if re.compile(script_regex[key]).match(text):
            return key

    return 'unknown'


# Unicameral conversion
def convert_unicameral(text, dir_from, dir_to):
    return ''.join(map(
        lambda letter : letter_converter(
            letter, 
            dir_from, 
            dir_to
        ),
        text
    ))


# Bicameral conversion of Georgian texts. 
# "'anbanismtavruli' is just a temp name for now "- said George ages ago. 
# TODO come up with a better name hopefully with historical precedent. 
def convert_bicameral(text, dir_from, dir_to):
    rules = {
        "anbanismtavruli": {
            "upper": "asomtavruli",
            "lower": "mtavruli",
        },
        "tfileliseuli": {
            "upper": "mtavruli",
            "lower": "mkhedruli",
        },
        "shanidziseuli": {
            "upper": "asomtavruli",
            "lower": "mkhedruli",
        },
        "khutsuri": {
            "upper": "asomtavruli",
            "lower": "nuskhuri",
        },
    }

    upper_script = rules[dir_to]['upper']
    lower_script = rules[dir_to]['lower']

    # Convert first letter
    converted = letter_converter(
        text[0], dir_from, upper_script
    ) + convert_unicameral(text[1:], dir_from, lower_script)

    # Convert first letter after every sentence
    matched = re.finditer(r'[?.!჻]\s+[A-zა-ჿႠ-Ⴥⴀ-ⴥᲐ-Ჿ0-9]', converted)
    converted = list(converted)
    for match in matched:
        converted[match.end()-1] = letter_converter(
            converted[match.end()-1], lower_script, upper_script
        )

    # Convert first letter before every special letter
    matched = re.finditer(r'[ა-ჿႠ-Ⴥⴀ-ⴥᲐ-Ჿ]\'', ''.join(converted))
    for match in matched:
        converted[match.start()] = letter_converter(
            converted[match.start()], lower_script, upper_script
        )
        converted[match.start()+1] = ''

    return ''.join(converted)


def letter_converter(letter, dir_from, dir_to):
    try: return data['alphabets'][dir_to][
        data['alphabets'][dir_from].index(letter)
    ]
    except: return letter



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='anbani',
    version='0.9.5',
    author="George Gach",
    author_email="georgegach@outlook.com",
    description="Georgian alphabet and language utilities for Natural Language Processing, script conversion and more.",
    license='GPL',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/anbani/anbani.py",
    include_package_data=True,
    packages=find_packages(),
    install_requires=[
        "pymupdf",
        "hjson",
        "numpy",
        "pandas",
        "tqdm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



import pandas as pd
from tqdm import tqdm
import re
from collections import defaultdict
import fitz

def ebook2text(path):
    try:
        doc = fitz.open(path)
        n_pages = doc.page_count
        
        text = ""
        
        for p in range(n_pages):
            text += doc.load_page(p).get_text()
        
        # Replace hyphen-newline to connect words
        text = text.replace('-\n', '') # 002D
        text = text.replace('‐\n', '') # 2010
        text = text.replace('‑\n', '') # 2011
        text = text.replace('‒\n', '') # 2012
        text = text.replace('–\n', '') # 2013
        text = text.replace('—\n', '') # 2014
        text = text.replace('―\n', '') # 2015
        text = text.replace('−\n', '') # 2212
        text = text.replace('֊\n', '') # 058A
        text = text.replace('⁻\n', '') # 207B
        text = text.replace('₋\n', '') # 208B
        text = text.replace('⸺\n', '') # 2E3A	
        text = text.replace('⸻\n', '') # 2E3B	
        text = text.replace('﹘\n', '') # FES8
        text = text.replace('－\n', '') # FF0D	
        text = text.replace('﹣\n', '') # FE63

        # Replace paragraph and random new lines
        text = text.replace('\n', ' ')
    except Exception as e:
        return 'Error: ' + str(e)

    return text
